(*----------------------------------------------------------------------------*
 *  Direct3D TD3D8GifImage class - load and render GIF file.                  *
 *  Features:                                                                 *
 *  1) GIF's are rendered by blending two nearest frames of GIF animation on  *
 *     multitexture capable hardware                                          *
 *  2) 3 different texture pipeline setups tested for best quality rendering  *
 *     result: (a) simplests NVIDIA TNT class, (b) Matrox G400 class,         *
 *     (c) NVIDIA GeForce, ATi Radeon - 3 stage setup.                        *
 *  3) Usage of Ander's Melander TGifImage component is reduced to loading    *
 *     stage - animation frames loaded to Direct3D textures, timing info      *
 *     retrieved to internal data structure. All rendering and timing after   *
 *     that is handled internally in class.                                   *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 03-Oct-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)
{******************************************************************************}
{                                                                              }
{ The contents of this file are used with permission, subject to the Mozilla   }
{ Public License Version 1.1 (the "License"); you may not use this file except }
{ in compliance with the License. You may obtain a copy of the License at      }
{ http://www.mozilla.org/MPL/MPL-1.1.html                                      }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}

unit D3D8GIF;

interface

{$Include DirectX.inc}

uses
  Windows, GifImage, Classes, SysUtils, Graphics, D3DX8,
  {$IFDEF DXG_COMPAT} DirectXGraphics {$ELSE} Direct3D8 {$ENDIF};

type
  TD3D8GifImage = class
    // Initializing and destroying device-dependent objects
    function InitDeviceObjects(pd3dDevice: IDirect3DDevice8): HRESULT;
    function RestoreDeviceObjects: HRESULT;
    function InvalidateDeviceObjects: HRESULT;
    function DeleteDeviceObjects: HRESULT;

    // Constructor / destructor
    constructor Create(aGifFileName: String; aReleaseGifComponent: Boolean{$IFDEF SUPPORTS_DEFAULTPARAMS} = False{$ENDIF});
    destructor Destroy; {$IFDEF TMT}virtual{$ELSE}override{$ENDIF};
  private
    // property related data
    fReleaseGifComponent: Boolean;
    fGifFileName: String;
    // internal data
    fGifImage: TGifImage;
    fDelays: array of Word; // in picoseconds (1/100 ?)
    fTotalLength: Word;
    fTime: DWORD;
    fPaused: Boolean;
    fPausedTime: DWORD;
    fIsBlendingSupported: Boolean;
    fIsDiffuseModulated: Boolean; // Is final blended GIF modulated by diffuse?
    fEnchancedFiltering: Boolean; // Use triliner with anisotropic filtering on textures
    // D3D Device related data
    fD3DDevice: IDirect3DDevice8;
    fTextures: array of IDirect3DTexture8;
    fSavedStateBlock, fDrawGifStateBlock: DWORD;
    fAnimationSpeed: DWORD; // 100 == Original
    // Info
    mFrameNum: DWORD;
    mBlendFactor: Single;
    procedure SetRenderStates(UseDiffuse: Boolean; EnchancedFiltering: Boolean);
    function ValidatePossibleMethods: Boolean;
    procedure SetGifFileName(const Value: String);
    procedure UploadBitmaps;
  public
    procedure StartTimer(DoRestart: Boolean = False);
    procedure StopTimer;
    function ApplyStates: Boolean;
    procedure RestoreStates;
    property ReleaseGifComponent: Boolean read fReleaseGifComponent write fReleaseGifComponent;
    property GifFileName: String read fGifFileName write SetGifFileName;
    property IsDiffuseModulated: Boolean read fIsDiffuseModulated;
    property EnchancedFiltering: Boolean read fEnchancedFiltering; 
    property AnimationSpeed: DWORD read fAnimationSpeed write fAnimationSpeed;
    property FrameNum: DWORD read mFrameNum;
    property BlendFactor: Single read mBlendFactor;
  end;

implementation

uses MMSystem;

{ TD3D8GifImage }

constructor TD3D8GifImage.Create(aGifFileName: String; aReleaseGifComponent: Boolean);
begin
  fIsBlendingSupported:= False;
  fIsDiffuseModulated:= False;
  fEnchancedFiltering:= False;
  fAnimationSpeed:= 100;

  fGifImage := TGIFImage.Create;
  fGifImage.DrawOptions:= GIFImageDefaultDrawOptions - [goTransparent, goAnimate];

  fReleaseGifComponent:= aReleaseGifComponent;
  GifFileName:= aGifFileName;

  timeBeginPeriod(5);
  StartTimer(True);
end;

destructor TD3D8GifImage.Destroy;
begin
  timeEndPeriod(5);

  InvalidateDeviceObjects;
  DeleteDeviceObjects;
  FreeAndNil(fGifImage);
  inherited Destroy;
end;

function TD3D8GifImage.InvalidateDeviceObjects: HRESULT;
begin
  // Delete the state blocks
  if (fD3DDevice <> nil) then
  begin
    if (fSavedStateBlock <> 0) then
      fD3DDevice.DeleteStateBlock(fSavedStateBlock);
    if (fDrawGifStateBlock <> 0) then
      fD3DDevice.DeleteStateBlock(fDrawGifStateBlock);
  end;

  fSavedStateBlock   := 0;
  fDrawGifStateBlock := 0;

  Result:= S_OK;
end;

function TD3D8GifImage.DeleteDeviceObjects: HRESULT;
var
  i: Integer;
begin
  for i:= 0 to High(fTextures) do fTextures[i]:= nil;
  fTextures:= nil; // Clean up dynamic array
  fDelays:= nil; // Clean up dynamic array
  fD3DDevice:= nil;
  Result:= S_OK;
end;

function TD3D8GifImage.InitDeviceObjects(
  pd3dDevice: IDirect3DDevice8): HRESULT;
begin
  fD3DDevice:= pd3dDevice;
  UploadBitmaps;
  Result:= S_OK;
end;

function TD3D8GifImage.RestoreDeviceObjects: HRESULT;
var
  which: LongWord;
begin
  // Determine if we could actually draw anything with current hardware
  fIsBlendingSupported:= ValidatePossibleMethods;

  // Create the state blocks for rendering GIF's
  for which:= 0 to 1 do
  begin
    fD3DDevice.BeginStateBlock;

    SetRenderStates(fIsDiffuseModulated, fEnchancedFiltering);

    if (which = 0) then
    begin // so original textures can be restored
      fD3DDevice.SetTexture(0, nil);
      fD3DDevice.SetTexture(1, nil);
    end;

    if (which = 0) then
      fD3DDevice.EndStateBlock(fSavedStateBlock)
    else
      fD3DDevice.EndStateBlock(fDrawGifStateBlock);
  end;

  Result:= S_OK;
end;

procedure TD3D8GifImage.SetRenderStates(UseDiffuse: Boolean; EnchancedFiltering: Boolean);
begin
  // This is simple mode for [possible] all two texture stage cards, like TNT
  fD3DDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  fD3DDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  fD3DDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  fD3DDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
  fD3DDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  fD3DDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  if EnchancedFiltering then
  begin
    fD3DDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_ANISOTROPIC);
    fD3DDevice.SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR); // Trilinear
  end else
  begin
    fD3DDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    fD3DDevice.SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_POINT);
  end;

  fD3DDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_BLENDFACTORALPHA);
  fD3DDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  fD3DDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
  fD3DDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
  fD3DDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 0); // Use the same first texture
  fD3DDevice.SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  if EnchancedFiltering then
  begin
    fD3DDevice.SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_ANISOTROPIC);
    fD3DDevice.SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_LINEAR); // Trilinear
  end else
  begin
    fD3DDevice.SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    fD3DDevice.SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_POINT);
  end;

  if UseDiffuse then
  begin
    // This is NVIDIA GeForce (1,2) special mode
    // - should be working on radeons throught
    fD3DDevice.SetTextureStageState(2, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
    fD3DDevice.SetTextureStageState(2, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    fD3DDevice.SetTextureStageState(2, D3DTSS_COLORARG2, D3DTA_CURRENT);
    fD3DDevice.SetTextureStageState(2, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    fD3DDevice.SetTextureStageState(3, D3DTSS_COLOROP,   D3DTOP_DISABLE);
    fD3DDevice.SetTextureStageState(3, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
  end else
  begin
    fD3DDevice.SetTextureStageState(2, D3DTSS_COLOROP,   D3DTOP_DISABLE);
    fD3DDevice.SetTextureStageState(2, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
  end;

  fD3DDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $FF000000);
end;

function TD3D8GifImage.ValidatePossibleMethods: Boolean;
var
  Passes: Cardinal;
begin
  if High(fTextures) < 1 then // if only single frame is available
  begin
    Result:= False;
    Exit;
  end;

  fD3DDevice.SetTexture(0, fTextures[0]);
  fD3DDevice.SetTexture(1, fTextures[1]);

  fIsDiffuseModulated:= True;
  fEnchancedFiltering:= True;

  Result:= True;

  SetRenderStates(IsDiffuseModulated, EnchancedFiltering);
  if Succeeded(fD3DDevice.ValidateDevice(Passes)) then Exit;

  fIsDiffuseModulated:= False;
  SetRenderStates(IsDiffuseModulated, EnchancedFiltering);
  if Succeeded(fD3DDevice.ValidateDevice(Passes)) then Exit;

  fEnchancedFiltering:= False;
  SetRenderStates(IsDiffuseModulated, EnchancedFiltering);
  if Succeeded(fD3DDevice.ValidateDevice(Passes)) then Exit;

  Result:= False;
end;

procedure TD3D8GifImage.SetGifFileName(const Value: String);
begin
  if (fGifFileName <> Value) then
  begin
    if not FileExists(Value) then
      raise Exception.Create('TD3D8GifImage: Can''t find GIF file');

    fGifImage.LoadFromFile(Value); // This could possibly raise exception...
    if Assigned(fD3DDevice) then
    begin
      UploadBitmaps;
      InvalidateDeviceObjects;
      if FAILED(RestoreDeviceObjects) then Exit;
        raise Exception.Create('TD3D8GifImage: ValidateDevice failed - ' +
                               'can''t support blended GIF output');
    end;

    fGifFileName:= Value;
  end;
end;

procedure TD3D8GifImage.UploadBitmaps;
var
  Bitmap: TBitmap;
  S: TMemoryStream;
  i: Integer;
begin
  if not Assigned(fD3DDevice) then Exit;

  Bitmap:= nil;
  S:= nil;
  try
    S:= TMemoryStream.Create;
    Bitmap:= TBitmap.Create;
    Bitmap.Assign(fGifImage);

    fTotalLength:= 0;
    SetLength(fDelays,   fGifImage.Images.Count);
    SetLength(fTextures, fGifImage.Images.Count);
    FillChar(fTextures[0], SizeOf(Pointer)*fGifImage.Images.Count, 0);

    for i:= 0 to fGifImage.Images.Count - 1 do
    begin
      S.Clear;

      fGifImage.Images[i].Draw(Bitmap.Canvas,
                               Rect(0, 0, Bitmap.Width, Bitmap.Height),
                               True, False);
      Bitmap.SaveToStream(S);
      
      D3DXCreateTextureFromFileInMemory(fD3DDevice, S.Memory^, S.Size, fTextures[i]);

      if (fGifImage.Images[i].GraphicControlExtension <> nil) then
        fDelays[i] := fGifImage.Images[i].GraphicControlExtension.Delay;
      if (fDelays[i] = 0) then fDelays[i]:= 1;
      Inc(fTotalLength, fDelays[i]); // Calculate total GIF animation length
    end;
  finally
    fGifImage.PaintStop;
    FreeAndNil(Bitmap);
    FreeAndNil(S);
  end;
end;

procedure TD3D8GifImage.StartTimer(DoRestart: Boolean);
var
  aNow: DWORD;
begin
  aNow:= timeGetTime;

  if DoRestart then fTime:= aNow;
  if fPaused then
  begin
    if (fPausedTime = 0) then fPausedTime:= aNow;
    fTime:= fTime - aNow + fPausedTime;
    fPaused:= False;
  end;
end;

procedure TD3D8GifImage.StopTimer;
begin
  if not fPaused then
  begin
    fPaused:= False;
    fPausedTime:= timeGetTime;
  end;
end;

function TD3D8GifImage.ApplyStates: Boolean;
var
  AnimTime, FrameStart,
  FrameNum, NextFrame: DWORD;
  BlendFactor: Single;
begin
  Result:= fIsBlendingSupported;
  if fIsBlendingSupported then
  begin
    // Setup renderstate
    fD3DDevice.CaptureStateBlock(fSavedStateBlock);
    fD3DDevice.ApplyStateBlock(fDrawGifStateBlock);

    // Calculate time passed from start of GIF animation
    AnimTime:= (((timeGetTime - fTime)*fAnimationSpeed) div (12*100)) mod fTotalLength;
    // Calculate what frame we are currently in
    FrameStart:= 0;
    FrameNum:= 0;
    while (FrameNum < DWORD(High(fDelays))) and
          (FrameStart + fDelays[FrameNum] < AnimTime) do
    begin
      Inc(FrameStart, fDelays[FrameNum]);
      Inc(FrameNum);
    end;

    if (FrameNum = DWORD(High(fDelays)))
    then NextFrame:= 0
    else NextFrame:= FrameNum + 1;

    BlendFactor:= (AnimTime - FrameStart) / fDelays[FrameNum];

    // Set blend constant
    fD3DDevice.SetTexture(0, fTextures[FrameNum]);
    fD3DDevice.SetTexture(1, fTextures[NextFrame]);
    fD3DDevice.SetRenderState(D3DRS_TEXTUREFACTOR,
      D3DCOLOR_COLORVALUE(BlendFactor, BlendFactor, BlendFactor, BlendFactor));

    mFrameNum:= FrameNum;
    mBlendFactor:= BlendFactor;
  end;
end;

procedure TD3D8GifImage.RestoreStates;
begin
  // Restore the modified renderstates
  fD3DDevice.ApplyStateBlock(fSavedStateBlock);
  fD3DDevice.SetTexture(0, nil);
  fD3DDevice.SetTexture(1, nil);
end;

end.

